unit xPidl;
// function for PIDL handling

interface

uses
  Windows, Messages,
  {$IFDEF VER100}ActiveX,{$ELSE}Ole2,{$ENDIF}
  ShlOBJ, xProcs;

type
  TShellRoot = (
    srDesktop, {srDesktopExpanded,} srPrograms, srControlPanel, srPrinters,
    srPersonal, srFavorites, srStartup, srRecent, srSendTo, srRecycleBin,
    srStartMenu, srDesktopDirectory, srDrives, srNetwork, srNetHood, srFonts,
    srTemplates  );

  TDisplayOption = (doInFolder, doForParsing);
  TDisplayOptions = set of TDisplayOption;

  TFolderOption = ( foFolders, foNonFolders, foHidden );
  TFolderOptions = set of TFolderOption;

{ TFolder }
type
  TFolder = class (TObject)
  private
    pidl,
    pidlfq   : PItemIDList;
    pFolder  : IShellFolder;
    FAttribs : uInt;
    function GetName: String;
    function GetParseName: String;
    function GetCount: Integer;
  public
    constructor Create(aFolder: IShellFolder);
    destructor Destroy; override;
    function  Compare(aFolder: TFolder): Integer;
    procedure ContextMenu(aHandle: hWnd; Where: TPoint);
    property Id : PItemIdList read pidl write pidl;
    property IdFq: PItemIdList read pidlfq write pidlfq;
    property Shell: IShellFolder read pFolder write pFolder;
    property Name: String read GetName;
    property ParseName: String read GetParseName;
    property Count: Integer read GetCount;
    property Attribs: uInt read FAttribs write FAttribs;
  end;

const
  CM_SHELLUPDATE = WM_USER + 101;

type
  TCMShellUpdate = record
    Msg: Cardinal;
    Unused: Integer;
    Sender: TFolder;
    Result: Longint;
  end;


// Retrieve next PIDL in list
function pidlNext(aPidl: PItemIdList): PItemIdList;

// Retrieve number of PIDL in list
function pidlCount(aPidl: PItemIdList): Integer;

// Retrieve complete size of complete PIDL
function pidlSize(aPidl: PItemIdList): Integer;

// Dump a pidl to debug terminal
procedure pidlDebug(aPidl: PItemIdList);

// create a new PIDL with a given size
function pidlCreate(aSize: Integer): PItemIdList;

// compare two PIDL's if they're equal
function pidlEqual(aPidl,bPidl: PItemIdList): Boolean;

// concat two pidls together
function pidlConcat(aId1,aId2: PItemIdList): PItemIdList;

// ???
function pidlCopyItemID(aPidl: PItemIdList): PItemIdList;

// get shell name for a PIDL
function pidlGetName(aFolder: IShellFolder;
                     aPidl: PItemIdList; aFlags: Integer): String;

// get specific attribs for a PIDL
function pidlAttribs(aPidl: PItemIdList; aFolder: IShellFolder): uInt;

// get fully qualified PIDL
function pidlGetQualified(aFolder: IShellFolder;
                          aId: PItemIdList): PItemIdList;

// get system icon index from PIDL
function pidlGetIconNo(aPIDL: PItemIdList; aOpen: Boolean): Integer;

// get system icon overlay index from PIDL
function pidlGetIconOverlay: Integer;

// get system image list
function pidlSysImageList(aSmall: Boolean): Integer;

var
  ShellMalloc : IMalloc;

const
  AddRefCount : Integer = 0;

procedure IncAdd;
procedure DecAdd;

implementation

uses
  SysUtils, ShellAPI;

const
  cbSize = Sizeof(Word);

//:::::::::::::::::::::::::::::::::::::::
procedure IncAdd;
begin
  Inc(AddRefCount);
//  strDebug('+'+IntToStr(AddRefCount));
end;

//:::::::::::::::::::::::::::::::::::::::
procedure DecAdd;
begin
  Dec(AddRefCount);
//  strDebug('-'+IntToStr(AddRefCount));
end;

{ TFolder }

//:::::::::::::::::::::::::::::::::::::::
constructor TFolder.Create(aFolder: IShellFolder);
begin
  inherited Create;
  pFolder:=aFolder;
  if Assigned(aFolder) then begin aFolder._AddRef; IncAdd; end;
end;

//:::::::::::::::::::::::::::::::::::::::
destructor TFolder.Destroy;
begin
  if Assigned(pidl)    then ShellMalloc.Free(pidl);
  if Assigned(pidlfq)  then ShellMalloc.Free(pidlfq);
  if Assigned(pFolder) then begin pFolder._Release; DecAdd; end;
  inherited Destroy;
end;

//:::::::::::::::::::::::::::::::::::::::
function TFolder.GetName: String;
begin
  Result := pidlGetName(pFolder, pidl, SHGDN_NORMAL);
end;

//:::::::::::::::::::::::::::::::::::::::
function TFolder.GetParseName: String;
begin
  try
    Result := pidlGetName(pFolder, pidl, SHGDN_FORPARSING);
  except
    Result := '';
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
function TFolder.GetCount: Integer;
begin
  Result := pidlCount(pidl);
end;

//:::::::::::::::::::::::::::::::::::::::
function TFolder.Compare(aFolder: TFolder): Integer;
var
  hr: hResult;
begin
  Result:=0;
  if Assigned(aFolder) then begin
    hr:=Shell.CompareIDs(0, Id, aFolder.Id);
    if Succeeded(hr) then Result:=SmallInt(ResultCode(hr));
  end;
end;

procedure TFolder.ContextMenu(aHandle: hWnd; Where: TPoint);
var
  aContextMenu : IContextMenu;
  aPidl        : PItemIDList;
  aPrgOut      : Pointer;
  aPopup       : hMenu;
  aCmd         : Integer;
  aCmdInfo     : TCMInvokeCommandInfo;
begin
  aPidl:=Id;
  if Succeeded(Shell.GetUIObjectOf(aHandle, 1, aPidl, IID_IContextMenu,
                                   aPrgOut, Pointer(aContextMenu))) then
    try
      aPopUp:=CreatePopUpMenu;
      try
        if Succeeded(aContextMenu.QueryContextMenu(aPopup, 0, 1, $7FFF,
                         CMF_DEFAULTONLY {CMF_NORMAL {CMF_EXPLORE})) then begin
          aCmd:=Integer(TrackPopupMenuEx(aPopup,
                        TPM_LEFTALIGN or TPM_RETURNCMD or TPM_RIGHTBUTTON or
                        TPM_HORIZONTAL or TPM_VERTICAL,
                        Where.X, Where.Y, aHandle, nil));
{
          aCmd:=Integer(TrackPopupMenu(aPopup,
                        TPM_LEFTALIGN or TPM_RETURNCMD or TPM_RIGHTBUTTON,
                        Where.X, Where.Y, 0, aHandle, nil));
}
          if aCmd <> 0 then begin // Execute command that was selected.
            Fillchar(aCmdInfo,Sizeof(aCmdInfo),0);
            with aCmdInfo do begin
              cbSize := SizeOf(TCMInvokeCommandInfo);
              hwnd   := aHandle;
              lpVerb := MakeIntResource(aCmd-1);
              nShow  := SW_SHOWNORMAL;
            end;
            if not Succeeded(aContextMenu.InvokeCommand(aCmdInfo)) then
              raise Exception.Create('not successfull command exec !');
          end;
        end;
       finally
         DestroyMenu(aPopup);
       end;
     finally
       aContextMenu._Release;
     end;
end;

{ functions }

//:::::::::::::::::::::::::::::::::::::::
function pidlNext(aPidl: PItemIdList): PItemIdList;
var
  cb: Integer;
begin
  Result:=nil;
  cb:=aPidl^.mkId.cb;
  if cb <> 0  then Result := PItemIdList(PChar(aPidl)+cb);
end;

function CompareMem(P1, P2: Pointer; Length: Integer): Boolean; assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     ESI,P1
        MOV     EDI,P2
        MOV     EDX,ECX
        XOR     EAX,EAX
        AND     EDX,3
        SHR     ECX,2
        REPE    CMPSD
        JNE     @@2
        MOV     ECX,EDX
        REPE    CMPSB
        JNE     @@2
@@1:    INC     EAX
@@2:    POP     EDI
        POP     ESI
end;


//:::::::::::::::::::::::::::::::::::::::
function pidlEqual(aPidl,bPidl: PItemIdList): Boolean;
begin
  Result:=False;
  if Assigned(aPidl) and Assigned(bPidl) then
     if aPidl^.mkId.cb=bPidl^.mkId.cb then
        Result:=CompareMem(@aPidl^.mkId.abID, @bPidl^.mkId.abID,
                           aPidl^.mkId.cb-cbSize);
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlCount(aPidl: PItemIdList): Integer;
begin
  Result:=0;
  while Assigned(aPidl) and (aPidl^.mkId.cb<>0) do begin
    Inc(Result);
    aPidl:=pidlNext(aPidl);
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
procedure pidlDebug(aPidl: PItemIdList);
var
  P : PChar;
  i : Integer;
begin
  if Assigned(aPidl) and (aPidl^.mkId.cb<>0) then
  begin
    P:=PChar(@aPidl^.mkid.abId);
    i:=aPidl^.mkId.cb-cbSize;
    while i>0 do
    begin
      if P^ in [' '..#127] then
         write(P^)
      else
         write('#',Ord(P^));
      write(' ');
      Inc(P);
      Dec(i);
    end;
    writeln;
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlSize(aPidl: PItemIdList): Integer;
begin
  Result:=0;
  if Assigned(aPidl) then begin
    Result := cbSize;
    while (aPidl^.mkid.cb <> 0) do begin
      Result:= Result + aPidl^.mkid.cb;
      aPidl:= pidlNext(aPidl);
    end;
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlCreate(aSize: Integer): PItemIdList;
begin
  Result:= ShellMalloc.Alloc(aSize);
  Fillchar(Result^,aSize,0);
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlConcat(aId1,aId2: PItemIdList): PItemIdList;
var
  aSize1,
  aSize2   : Integer;
  aPidlNew : PItemIdList;
begin
  if Assigned(aId1) then aSize1 := pidlSize(aId1) - cbSize
  else aSize1 := 0;

  aSize2:= pidlSize(aId2);

  aPidlNew:= pidlCreate(aSize1+aSize2);
  if Assigned(aPidlNew) then begin
    if Assigned(aId1) then Move(aId1^,aPidlNew^, aSize1);
    Move(aId2^, (PChar(aPidlNew)+aSize1)^, aSize2);
  end;
  Result:=aPidlNew;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlCopyItemID(aPidl: PItemIdList): PItemIdList;
begin
  with aPidl^.mkId do begin
    Result:= ShellMalloc.Alloc(cb+cbSize);
    CopyMemory(Result, aPidl, cb+cbSize);
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlAttribs(aPidl: PItemIdList; aFolder: IShellFolder): uInt;
begin
  Result:= SFGAO_HASSUBFOLDER or SFGAO_FOLDER or SFGAO_LINK or SFGAO_SHARE;
  aFolder.GetAttributesOf(1, aPidl, Result);
end;

// ok
//:::::::::::::::::::::::::::::::::::::::
function pidlGetName(aFolder: IShellFolder;
                     aPidl: PItemIdList; aFlags: Integer): String;
var
  aName : TStrRet;
  aInfo : array[0..500] of char;
begin
  Result:='';
  if Succeeded(aFolder.GetDisplayNameOf(aPidl, aFlags, aName)) then
    case aName.uType of
      STRRET_WSTR :
        begin
          WideCharToMultiByte(CP_ACP, 0, aName.pOleStr,
                               -1, aInfo, Sizeof(aInfo)-1, nil, nil);
          Result:=String(aInfo);
        end;

      STRRET_OFFSET:
        Result:=PChar(longint(aPidl)+aName.uOffset);

      STRRET_CSTR:
        Result:=aName.cStr;
    end;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlGetQualified(aFolder: IShellFolder;
                          aId: PItemIdList): PItemIdList;
var
  aName    : String;
  aInfo    : array[0..MAX_PATH] of WideChar;
  aDesktop : IShellFolder;
  aPidl    : PItemIdList;
  aEaten,
  aAttribs : Longint;
begin
  Result := nil;
  aName:=pidlGetName(aFolder, aId, SHGDN_FORPARSING);
  if aName<>'' then begin
    if Succeeded(SHGetDesktopFolder(aDesktop)) then
    try
      MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED,
                          PChar(aName), -1, aInfo, Sizeof(aInfo));
      if Succeeded(aDesktop.ParseDisplayName(0,
                            nil, aInfo, aEaten, aPidl, aAttribs)) then
        Result := aPIdl;
    finally
      aDesktop._Release;
    end;
  end;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlGetIconNo(aPIDL: PItemIdList; aOpen: Boolean): Integer;
var
  aInfo     : TSHFileInfo;
  aFlag     : uInt;
begin
  aFlag := SHGFI_PIDL or SHGFI_SYSICONINDEX;
  if aOpen then aFlag := aFlag or SHGFI_OPENICON;
  SHGetFileInfo(PChar(aPIDL), 0, aInfo, Sizeof(aInfo), aFlag );
  Result:=aInfo.iIcon;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlGetIconOverlay: Integer;
var
  aInfo     : TSHFileInfo;
  aFlag     : uInt;
begin
  aFlag := SHGFI_SYSICONINDEX or SHGFI_SMALLICON or SHGFI_LINKOVERLAY;
  SHGetFileInfo(nil, 0, aInfo, Sizeof(aInfo), aFlag );
  Result:=aInfo.iIcon;
end;

//:::::::::::::::::::::::::::::::::::::::
function pidlSysImageList(aSmall: Boolean): Integer;
var
  aInfo: TSHFileInfo;
  aFlag: uInt;
begin
  if aSmall then aFlag := SHGFI_SMALLICON else aFlag := SHGFI_LARGEICON;
  Result:= SHGetFileInfo('', 0, aInfo,
                         Sizeof(aInfo), SHGFI_SYSICONINDEX or aFlag );
end;

initialization

  if SHGetMalloc(ShellMalloc) <> NOERROR then
     Information('Error getting Shell memory allocator.');

finalization

  ShellMalloc._Release;

end.


end.




