unit xShlList;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ShlObj, xPidl;

{ TcomShellListView }
type
  TxShellListView = class(TCustomListView)
  private
    { Private-Deklarationen }
    FShellRoot      : TShellRoot;
    FDisplayOptions : TDisplayOptions;
    FFolderOptions  : TFolderOptions;
    FDispOption,
    FFoldOption     : uInt;
    FSmall,
    FLarge          : TImageList;
    FSetting        : TFolder;
    procedure BuildList;
    procedure FillListView(aFolder: IShellFolder; aPidlFq: PItemIdList);
    procedure CMShellUpdate(var Message: TCMShellUpdate); message CM_SHELLUPDATE;
    procedure SetShellRoot(Value: TShellRoot);
    procedure SetDisplayOptions(Value: TDisplayOptions);
    procedure SetFolderOptions(Value: TFolderOptions);
    function  AddItem(aPidl, aPidlfq: PItemIdList; aFolder: IShellFolder): TListItem;
    procedure Sort;
    procedure ClearList;
    function  GetPidl: PItemIdList;
    function  IsMaskStored: Boolean;
    function  IsMaskAllFiles: Boolean;
    procedure SetMask(const NewMask: string);
  protected
    { Protected-Deklarationen }
    FMask: string;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    { Public-Deklarationen }
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure ResetList(aData: TFolder);
    property Id: PItemIdList read GetPidl;
  published
    { Published-Deklarationen }
    property ShellRoot: TShellRoot read FShellRoot
              write SetShellRoot default srDesktop;
    property DisplayOptions: TDisplayOptions read FDisplayOptions
              write SetDisplayOptions;
    property FolderOptions: TFolderOptions read FFolderOptions
              write SetFolderOptions;
    property Mask: string read FMask write SetMask stored IsMaskStored;
    property Align;
    property BorderStyle;
    property Color;
    property ColumnClick;
    property OnClick;
    property OnDblClick;
    property Columns;
    property Ctl3D;
    property DragMode;
    property ReadOnly;
    property Font;
    property HideSelection;
    property IconOptions;
    property AllocBy;
    property MultiSelect;
    property OnChange;
    property OnChanging;
    property OnColumnClick;
    property OnCompare;
    property OnDeletion;
    property OnEdited;
    property OnEditing;
    property OnEnter;
    property OnExit;
    property OnInsert;
    property OnDragDrop;
    property OnDragOver;
    property DragCursor;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property ParentShowHint;
    property ShowHint;
    property PopupMenu;
    property ShowColumnHeaders;
    property SortType;
    property TabOrder;
    property TabStop default True;
    property ViewStyle;
    property Visible;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
  end;

procedure Register;

implementation

uses
  ShellAPI,ActiveX,xProcs,DsgnIntf;

const
  DefaultMask = '*.*';

  RootValues: array[TShellRoot] of uInt = (
    CSIDL_DESKTOP, { $0001,} CSIDL_PROGRAMS, CSIDL_CONTROLS, CSIDL_PRINTERS,
    CSIDL_PERSONAL, CSIDL_FAVORITES, CSIDL_STARTUP, CSIDL_RECENT, CSIDL_SENDTO,
    CSIDL_BITBUCKET, CSIDL_STARTMENU, CSIDL_DESKTOPDIRECTORY, CSIDL_DRIVES,
    CSIDL_NETWORK, CSIDL_NETHOOD, CSIDL_FONTS, CSIDL_TEMPLATES
   );

constructor TxShellListView.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FSmall:=TImageList.CreateSize(16,16);
  FSmall.ShareImages:=True;
  FSmall.Handle:= pidlSysImageList(True);
  SmallImages:=FSmall;

  FLarge:=TImageList.CreateSize(32,32);
  FLarge.ShareImages:=True;
  FLarge.Handle:= pidlSysImageList(False);
  LargeImages:=FLarge;

  FShellRoot      := srDesktop;
  FDispOption     := SHGDN_NORMAL;
  FFoldOption     := SHCONTF_NONFOLDERS;
  FFolderOptions  := [foFolders];

  FMask := DefaultMask;  { default file mask is all }
end;

destructor TxShellListView.Destroy;
begin
  FSmall.Free;
  FLarge.Free;
  inherited Destroy;
end;

procedure TxShellListView.SetShellRoot(Value: TShellRoot);
begin
  if Value <> FShellRoot then begin
    FShellRoot:=Value;
    RecreateWnd;
  end;
end;

procedure TxShellListView.SetDisplayOptions(Value: TDisplayOptions);

  function ConvertDisplay: uInt;
  const
    FlagValues: array[TDisplayOption] of UINT = (
      SHGDN_INFOLDER, SHGDN_FORPARSING );
  var
    Opt: TDisplayOption;
  begin
    Result :=SHGDN_NORMAL;
    for Opt := Low(TDisplayOption) to High(TDisplayOption) do
      if Opt in FDisplayOptions then
        Result := Result OR FlagValues[Opt];
  end;

begin
  if Value <> FDisplayOptions then begin
    FDisplayOptions:=Value;
    FDispOption := ConvertDisplay;
    RecreateWnd;
  end;
end;

procedure TxShellListView.SetFolderOptions(Value: TFolderOptions);

  function Convert: uInt;
  const
    Values: array[TFolderOption] of UINT = (
      SHCONTF_FOLDERS,  SHCONTF_NONFOLDERS, SHCONTF_INCLUDEHIDDEN  );
  var
    Opt: TFolderOption;
  begin
    Result :=SHGDN_NORMAL;
    for Opt := Low(Opt) to High(Opt) do
      if Opt in FFolderOptions then
        Result := Result OR Values[Opt];
  end;

begin
  if Value <> FFolderOptions then begin
    FFolderOptions:=Value;
    FFoldOption := Convert;
    RecreateWnd;
  end;
end;

function TxShellListView.AddItem(
    aPidl, aPidlfq: PItemIdList; aFolder: IShellFolder): TListItem;
var
  aData   : TFolder;
  aFile,
  aExt    : String;
begin
  aData := TFolder.Create(aFolder);
  with aData do begin
    Id := pidlCopyItemID(aPidl);

    if not IsMaskAllFiles then begin
      aFile := ParseName;
      aExt := '*'+strLower(ExtractFileExt(aFile))+';';
      if Pos(aExt, FMask) = 0 then exit;
    end;

    IdFq := pidlConcat(aPidlfq, aPidl);
    Attribs := pidlAttribs(aPidl, aFolder);
    // Add node for current item
    Result:= Items.Insert(0);
    with Result do begin
      Result.Caption:=pidlGetName(aFolder,aPidl,FDispOption);
      Result.SubItems.Add(Caption);
      Result.Data := aData;
      // Set icons
      ImageIndex   := pidlGetIconNo(IdFq, False);
      if Attribs and SFGAO_LINK <> 0 then OverlayIndex := 1
      else if Attribs and SFGAO_SHARE <> 0 then OverlayIndex := 0; {2 Powertoys!}
    end;
  end;
end;

function TxShellListView.GetPidl: PItemIdList;
begin
  Result:=nil;
  if Selected <> nil then
     if Selected.Data <> nil then
        Result := TFolder(Selected.Data).Id;
end;

procedure TxShellListView.FillListView(aFolder: IShellFolder; aPidlFq: PItemIdList);
var
  aEnum     : IEnumIdList;
  aCount    : Longint;
  pidl      : PItemIdList;
  aTime: TDateTime;
  h,m,s,ms: Word;
begin
  Cursor := crHourGlass;
  // Turn off updating the tree
  Items.BeginUpdate;
  try
    ClearList;
    // Get the enumerator for the given folder
    if S_OK = aFolder.EnumObjects(0, FFoldOption, aEnum ) then begin
      // Loop over the list of objects
      while S_OK = aEnum.Next(1, pidl, aCount) do
        AddItem(pidl, aPidlFq, aFolder);
{
      SortType:=stText;
      aTime := Now;
      strDebug('Now into sort !');
      Sort;
      aTime := Now-aTime;
      DecodeTime(aTime, h,m,s,ms);
      strDebug(IntToStr(s*1000+ms));
      strDebug('Now out of sort !');
}
    end;
  finally
    // Refresh the tree
    Items.EndUpdate;
    Cursor := crDefault;
  end;
end;

procedure TxShellListView.Sort;

var
  T: TListItem;

  function Compare(a1,a2: TFolder): Integer;
  begin
    if a1.Name = a2.Name then      Result := 0
    else if a1.Name > a2.Name then Result := 1
    else                           Result := -1;
  end;

  procedure QuickSort(iLo, iHi: Integer);
  var
    Lo, Hi  : Integer;
    Mid     : TFolder;
  begin
    with Items do
      repeat
        Lo  := iLo;
        Hi  := iHi;
        Mid := TFolder(Item[(iLo + iHi) shr 1].Data);
        repeat
          while Compare(TFolder(Item[Lo].Data),Mid) < 0 do Inc(Lo);
          while Compare(TFolder(Item[Hi].Data),Mid) > 0 do Dec(Hi);
          if Lo <= Hi then begin
            if Lo <> Hi then begin
              T.Assign(Item[Lo]);
              Item[Lo].Assign(Item[Hi]);
              Item[Hi].Assign(T);
            end;
            Inc(Lo);
            Dec(Hi);
          end;
        until Lo > Hi;
        if iLo < Hi then QuickSort(iLo, Hi);
        iLo := Lo;
      until Lo >= iHi;
  end;

begin
  with Items do if Count > 1 then begin
    T:=TListItem.Create(Items);
    try
      QuickSort(0, Count-1);
    finally
      T.Free;
    end;
  end;
end;

procedure TxShellListView.ResetList(aData: TFolder);
var
  aFolder : IShellFolder;
begin
  if Assigned(aData) then
    with aData do
      if Succeeded(Shell.BindToObject(Id,nil,IID_IShellFolder,Pointer(aFolder))) then
         FillListView(aFolder, IdFq)
      else
         FillListView(Shell,Id);
  FSetting:=aData;
end;

procedure TxShellListView.CMShellUpdate(var Message: TCMShellUpdate);
begin
  ResetList(Message.Sender);
end;

procedure TxShellListView.BuildList;
var
  aRoot     : PItemIDList;
  aFolder   : IShellFolder;
  aData     : TFolder;
begin
  if not Succeeded(SHGetDesktopFolder(aFolder)) then
     raise Exception.Create('No desktop folder');

  if not Succeeded(SHGetSpecialFolderLocation(0, RootValues[FShellRoot], aRoot)) then
     raise Exception.Create('Error getting SHGetSpecialFolderLocation.');

  aData:=TFolder.Create(aFolder);
  with aData do begin
    Attribs:= pidlAttribs(aRoot, aFolder);
    Id     := aRoot;
    IdFq   := aRoot;
  end;
  ResetList(aData);
end;

procedure TxShellListView.ClearList;
var i : Integer;
begin
  for i:= 0 to Items.Count-1 do
    if Items[i].Data <> nil then TFolder(Items[i].Data).Free;
  Items.Clear;
end;

procedure TxShellListView.CreateWnd;
begin
  inherited CreateWnd;
  BuildList;
end;

procedure TxShellListView.DestroyWnd;
begin
  ClearList;
  inherited DestroyWnd;
end;

function TxShellListView.IsMaskStored: Boolean;
begin
  Result := DefaultMask <> FMask;
end;

function TxShellListView.IsMaskAllFiles: Boolean;
begin
  Result := Pos(DefaultMask, FMask)<>0;
end;

procedure TxShellListView.SetMask(const NewMask: string);
begin
//      if Pos(aExt, FMask) = 0 then exit;
//  if Pos(DefaultMask, NewMask) <> 0 then begin
//  end;
  if FMask <> NewMask then begin
    FMask := NewMask;
    ClearList;
    ResetList(FSetting);
  end;
end;

procedure TxShellListView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  aItem: TListItem;
begin
  inherited MouseDown(Button, Shift, X, Y);
  if ssRight in Shift then
  begin
    aItem:= GetItemAt(X, Y);
    if Assigned(aItem) then
    begin
      Selected:=aItem;
      TFolder(aItem.Data).ContextMenu(Handle, ClientToScreen(Point(X, Y)));
    end;
  end;
end;

procedure Register;
begin
  RegisterComponents('xTool', [TxShellListView]);
end;

initialization

end.
