unit xShlTree;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ShlOBJ, xPidl;

{ TcomShellTreeView }
type
  TxShellTreeView = class(TCustomTreeView)
  private
    { Private-Deklarationen }
    FShellRoot      : TShellRoot;
    FDisplayOptions : TDisplayOptions;
    FFolderOptions  : TFolderOptions;
    FDispOption,
    FFoldOption     : uInt;
    FSmall,
    FLarge          : TImageList;
    FListView       : TCustomListView;
    FCached         : Boolean;
    FMouseNode      : TTreeNode;
    fSmallImages    : Boolean;
    procedure BuildTree;
    procedure FillTreeView(aParent: TTreeNode;
              aFolder: IShellFolder; pidlfq: PItemIdList);
    procedure SetShellRoot(Value: TShellRoot);
    procedure SetDisplayOptions(Value: TDisplayOptions);
    procedure SetFolderOptions(Value: TFolderOptions);
    function  AddItem(aParent: TTreeNode; aPidl, aPidlfq: PItemIdList;
              aFolder: IShellFolder): TTreeNode;
    procedure DoDeletion(Sender: TObject; Node: TTreeNode);
    function  GetPidl: PItemIdList;
    procedure Sort(aParent: TTreeNode);
    procedure SetDirectory(Value: String);
    function  GetDirectory: String;
    procedure SetSmallImages(Value: Boolean);
    procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
  protected
    { Protected-Deklarationen }
    function  CanCollapse(Node: TTreeNode): Boolean; override;
    function  CanExpand(Node: TTreeNode): Boolean; override;
    procedure Change(Node: TTreeNode); override;
    procedure CreateWnd; override;
    procedure DestroyWnd; override;
    procedure Notification(AComponent: TComponent;
                                            Operation: TOperation); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
                                                      X,Y: Integer); override;
  public
    { Public-Deklarationen }
    constructor Create(aOwner: TComponent); override;
    destructor  Destroy; override;
    property Id: PItemIdList read GetPidl;
    property Directory: String read GetDirectory write SetDirectory;
  published
    { Published-Deklarationen }
    property ShellRoot: TShellRoot read FShellRoot
              write SetShellRoot default srDesktop;
    property SmallImages: Boolean read FSmallImages write SetSmallImages default True;
    property DisplayOptions: TDisplayOptions read FDisplayOptions
              write SetDisplayOptions;
    property FolderOptions: TFolderOptions read FFolderOptions
              write SetFolderOptions;
    property ListView: TCustomListView read FListView write FListView;
    property Cached: Boolean read FCached write FCached default True;
//the following are not tested
    property ShowButtons;
    property BorderStyle;
    property DragCursor;
    property ShowLines;
    property ShowRoot;
    property ReadOnly;
    property DragMode;
    property HideSelection;
    property Indent;
//    property Items;
    property OnEditing;
    property OnEdited;
    property OnExpanding;
    property OnExpanded;
    property OnCollapsing;
    property OnCompare;
    property OnCollapsed;
    property OnChanging;
    property OnChange;
    property OnDeletion;
    property OnGetImageIndex;
    property OnGetSelectedIndex;
    property Align;
    property Enabled;
    property Font;
    property Color;
    property ParentColor;
    property ParentCtl3D;
    property Ctl3D;
//    property SortType;
    property TabOrder;
    property TabStop default True;
    property Visible;
    property OnClick;
    property OnEnter;
    property OnExit;
    property OnDragDrop;
    property OnDragOver;
    property OnStartDrag;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnDblClick;
    property OnKeyDown;
    property OnKeyPress;
    property OnKeyUp;
    property PopupMenu;
    property ParentFont;
    property ParentShowHint;
    property ShowHint;
//    property Images;
//    property StateImages;


  end;

procedure Register;

implementation

uses
  ShellAPI,ActiveX,xProcs,DsgnIntf;

const
  RootValues: array[TShellRoot] of uInt = (
    CSIDL_DESKTOP, { $0001,} CSIDL_PROGRAMS, CSIDL_CONTROLS, CSIDL_PRINTERS,
    CSIDL_PERSONAL, CSIDL_FAVORITES, CSIDL_STARTUP, CSIDL_RECENT, CSIDL_SENDTO,
    CSIDL_BITBUCKET, CSIDL_STARTMENU, CSIDL_DESKTOPDIRECTORY, CSIDL_DRIVES,
    CSIDL_NETWORK, CSIDL_NETHOOD, CSIDL_FONTS, CSIDL_TEMPLATES
   );

{ TxShellTreeView }

constructor TxShellTreeView.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
{
  FImageList:=TImageList.CreateSize(32,32);
  FImageList.ShareImages:=True;
  FImageList.Handle:= pidlSysImageList(False);
  Images:= FImageList;
}

  FSmall:=TImageList.CreateSize(16,16);
  FSmall.ShareImages:=True;
  FSmall.Handle:= pidlSysImageList(True);

  FLarge:=TImageList.CreateSize(32,32);
  FLarge.ShareImages:=True;
  FLarge.Handle:= pidlSysImageList(False);
  Images:=FSmall;

  FSmallImages    := True;
  FShellRoot      := srDesktop;
  FDispOption     := SHGDN_NORMAL;
  FFoldOption     := SHCONTF_FOLDERS;
  FFolderOptions  := [foFolders];
  ShowRoot        := False;
  OnDeletion      := DoDeletion;
  FCached         := True;
end;

destructor TxShellTreeView.Destroy;
begin
  FSmall.Free;
  FLarge.Free;
  inherited Destroy;
end;

procedure TxShellTreeView.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (aComponent = FListView) then
    FListView:= nil;
end;

procedure TxShellTreeView.BuildTree;
var
  aRoot     : PItemIDList;
  aFolder   : IShellFolder;
begin
  if not Succeeded(SHGetDesktopFolder(aFolder)) then
    raise Exception.Create('No desktop folder');

  if not Succeeded(SHGetSpecialFolderLocation(0, RootValues[FShellRoot], aRoot)) then
    raise Exception.Create('Error getting SHGetSpecialFolderLocation.');

  with AddItem(nil, aRoot, aRoot, aFolder) do Expanded := True;
end;

function TxShellTreeView.AddItem(aParent: TTreeNode;
    aPidl, aPidlfq: PItemIdList; aFolder: IShellFolder): TTreeNode;
var
  aData   : TFolder;
begin
  // Make a record with the folder interface and the PIDL so we can
  // expand the list if necessary
  aData := TFolder.Create(aFolder);
  // pidl handling
  with aData do begin
    if Assigned(aParent) then begin
      Id   := pidlCopyItemID(aPidl);
      IdFq := pidlConcat(aPidlfq, aPidl);
    end else begin
      Id   := aPidl;
      IdFq := aPidlfq;
    end;
    // Add node for current item
    Result  := Items.AddChildObjectFirst(aParent,
                        pidlGetName(aFolder, aPidl, FDispOption), aData);
    Attribs := pidlAttribs(aPidl, aFolder);
    with Result do begin
      HasChildren := Attribs and SFGAO_HASSUBFOLDER <> 0;
      // Set icons
      ImageIndex   := pidlGetIconNo(IdFq, False);
      if Attribs and SFGAO_FOLDER <> 0 then
        SelectedIndex := pidlGetIconNo(IdFq, True)
      else
        SelectedIndex := ImageIndex;
      if Attribs and SFGAO_LINK <> 0 then OverlayIndex := 1
      else if Attribs and SFGAO_SHARE <> 0 then OverlayIndex := 0; {2 Powertoys}
    end;
  end;
end;

procedure TxShellTreeView.FillTreeView(aParent: TTreeNode;
      aFolder: IShellFolder; pidlfq: PItemIdList);
var
  aEnum     : IEnumIdList;
  aCount    : Longint;
  pidl      : PItemIdList;
begin
  Cursor := crHourGlass;
  // Turn off updating the tree
  Items.BeginUpdate;
  try
    // Get the enumerator for the given folder
    if S_OK = aFolder.EnumObjects(0, FFoldOption, aEnum ) then begin
      // Loop over the list of objects
      while S_OK = aEnum.Next(1, pidl, aCount) do
        AddItem(aParent, pidl, pidlfq, aFolder);
      Sort(aParent);
    end;
  finally
    // Refresh the tree
    Items.EndUpdate;
    Cursor := crDefault;
  end;
end;

procedure TxShellTreeView.DoDeletion(Sender: TObject; Node: TTreeNode);
begin
  if Assigned(Node.Data) then TObject(Node.Data).Free;
end;

procedure TxShellTreeView.Sort(aParent: TTreeNode);

  procedure MyAssign(a,b: TTreeNode);
  begin
    a.Assign(b);
    if (TFolder(b.Data).Attribs and SFGAO_HASSUBFOLDER) <> 0 then
      a.HasChildren := True;
  end;

  procedure QuickSort(iLo, iHi: Integer);
  var
    Lo, Hi  : Integer;
    T       : TTreeNode;
    Mid     : TFolder;
  begin
    with aParent do
      repeat
        Lo  := iLo;
        Hi  := iHi;
        Mid := TFolder(Item[(iLo + iHi) shr 1].Data);
        repeat
          while TFolder(Item[Lo].Data).Compare(Mid) < 0 do Inc(Lo);
          while TFolder(Item[Hi].Data).Compare(Mid) > 0 do Dec(Hi);
          if Lo <= Hi then begin
            if Lo <> Hi then begin
              T:=TTreeNode.Create(Items);
              MyAssign(t,Item[Lo]);
              MyAssign(Item[Lo],Item[Hi]);
              MyAssign(Item[Hi],T);
              T.Free;
            end;
            Inc(Lo);
            Dec(Hi);
          end;
        until Lo > Hi;
        if iLo < Hi then QuickSort(iLo, Hi);
        iLo := Lo;
      until Lo >= iHi;
  end;

begin
  with aParent do if HasChildren and (Count > 1) then QuickSort(0, Count-1);
end;

function TxShellTreeView.CanCollapse(Node: TTreeNode): Boolean;
begin
  Result := inherited CanCollapse(Node);
  if Result then begin
    Cursor := crHourGlass;
    try
      Items.BeginUpdate;
      if not FCached then Node.DeleteChildren;
      Node.HasChildren := True;
      Items.EndUpdate;
    finally
      Cursor := crDefault;
    end;
  end;
end;

function TxShellTreeView.CanExpand(Node: TTreeNode): Boolean;
var
  aFolder : IShellFolder;
begin
  Result := inherited CanExpand(Node);
  if (csDesigning in ComponentState) and (Node.Parent <> nil) then exit;
  if Result and Assigned(Node.Data) then
     with TFolder(Node.Data) do
     if not FCached or (Node.Count = 0) then
       if Succeeded(Shell.BindToObject(Id,nil,IID_IShellFolder,Pointer(aFolder))) then
         FillTreeView(Node, aFolder, IdFq)
       else
         FillTreeView(Node, Shell, Id);
end;

procedure TxShellTreeView.Change(Node: TTreeNode);
begin
  inherited Change(Node);
  if Assigned(FListView) and Assigned(Node.Data) then begin
     FListView.Perform(CM_SHELLUPDATE,0,Longint(Node.Data));
  end;
end;

procedure TxShellTreeView.WMRButtonDown(var Message: TWMRButtonDown);
begin
  FMouseNode := GetNodeAt(Message.XPos, Message.YPos);
  if FMouseNode <> nil then
    with FMouseNode do begin
      Focused := True;
      Selected := True;
    end;
  inherited;
end;

procedure TxShellTreeView.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if ssRight in Shift then
    if Assigned(FMouseNode) then
      TFolder(FMouseNode.Data).ContextMenu(Handle, ClientToScreen(Point(X, Y)));
  inherited MouseDown(Button, Shift, X, Y);
end;

procedure TxShellTreeView.CreateWnd;
begin
  inherited CreateWnd;
  BuildTree;
end;

procedure TxShellTreeView.DestroyWnd;
begin
  Items.Clear;
  inherited DestroyWnd;
end;

procedure TxShellTreeView.SetSmallImages(VAlue: Boolean);
begin
  if Value<>fSmallImages then
  begin
    fSmallImages:=Value;
    if Value then Images:=FSmall else Images:=FLarge;
  end;
end;

procedure TxShellTreeView.SetShellRoot(Value: TShellRoot);
begin
  if Value <> FShellRoot then begin
    FShellRoot:=Value;
    RecreateWnd;
  end;
end;

procedure TxShellTreeView.SetDisplayOptions(Value: TDisplayOptions);

  function ConvertDisplay: uInt;
  const
    FlagValues: array[TDisplayOption] of UINT = (
      SHGDN_INFOLDER, SHGDN_FORPARSING );
  var
    Opt: TDisplayOption;
  begin
    Result :=SHGDN_NORMAL;
    for Opt := Low(TDisplayOption) to High(TDisplayOption) do
      if Opt in FDisplayOptions then
        Result := Result OR FlagValues[Opt];
  end;

begin
  if Value <> FDisplayOptions then begin
    FDisplayOptions:=Value;
    FDispOption := ConvertDisplay;
    RecreateWnd;
  end;
end;

procedure TxShellTreeView.SetFolderOptions(Value: TFolderOptions);

  function Convert: uInt;
  const
    Values: array[TFolderOption] of UINT = (
      SHCONTF_FOLDERS,  SHCONTF_NONFOLDERS, SHCONTF_INCLUDEHIDDEN  );
  var
    Opt: TFolderOption;
  begin
    Result :=SHGDN_NORMAL;
    for Opt := Low(Opt) to High(Opt) do
      if Opt in FFolderOptions then
        Result := Result OR Values[Opt];
  end;

begin
  if Value <> FFolderOptions then begin
    FFolderOptions:=Value;
    FFoldOption := Convert;
    RecreateWnd;
  end;
end;

function TxShellTreeView.GetPidl: PItemIdList;
begin
  Result:=nil;
  if Selected <> nil then
    if Selected.Data <> nil then
      Result := TFolder(Selected.Data).Id;
end;

procedure TxShellTreeView.SetDirectory(Value: String);
var
  aRoot    : TFolder;
  aInfo    : array[0..MAX_PATH] of WideChar;
  aPidl,tmp: PItemIdList;
  aEaten,
  aAttribs : Longint;

(******  procedure Dump(aPidl: PItemIdList);
  var
    P: Pointer;
  begin
    while Assigned(aPidl) and (aPidl^.mkId.cb<>0) do begin
      P := @(aPidl^.mkId.abId);
      if not Assigned(P) then Information('x');
      aPidl:=pidlNext(aPidl);
    end;
  end; ***)

  procedure ScanFor(aNode: TTreeNode; aPidl: PItemIdList);
  var
    i : Integer;
    P : TTreeNode;
  begin
    if Assigned(aPidl) then
    begin
      if pidlNext(aPidl)<>nil then aNode.Expanded:=True;
      Selected:=aNode;
      for i:=0 to aNode.Count-1 do
      begin
        P:=aNode.Item[i];

        pidlDebug(aPidl);
        pidlDebug(TFolder(P.Data).Id);

        if pidlEqual(aPidl, TFolder(P.Data).Id) then
        begin
          aPidl:=pidlNext(aPidl);
writeln('=');
pidlDebug(aPidl);
// pidlGetQualified(aFolder: IShellFolder; aId: PItemIdList): PItemIdList;

          ScanFor(P,aPidl);
          break;
        end else
writeln('<> ');
      end;
    end;
  end;

begin
  aRoot:=TFolder(TopItem.Data);
  MultiByteToWideChar(CP_ACP, MB_PRECOMPOSED,
                      PChar(Value), -1, aInfo, Sizeof(aInfo));
  if Succeeded(aRoot.Shell.ParseDisplayName(0,nil,aInfo,aEaten,aPidl,aAttribs)) then
  begin

  // check complete pidl path
    tmp:=aPidl;
    while Assigned(tmp) and (tmp^.mkId.cb<>0) do begin
      writeln(pidlGetName(aRoot.Shell, tmp, SHGDN_FORPARSING));
      tmp:=pidlNext(tmp);
    end;
(*
    Cursor:=crHourGlass;
  //  Items.BeginUpdate;
    try
      ScanFor(TopItem, aPidl);
    finally
//      Items.EndUpdate;
      Cursor := crDefault;
    end;  *)
    ShellMalloc.Free(aPidl);
  end;
  { Value to Pidl }
  { Compare to selected pidl  BindToObject }
  { while pidl do Open ; next pidl }
end;

function  TxShellTreeView.GetDirectory: String;
begin
  Result:='';
  if Selected <> nil then
     if Selected.Data <> nil then
        Result := TFolder(Selected.Data).ParseName;
end;

procedure Register;
begin
  RegisterComponents('xTool', [TxShellTreeView]);
end;

initialization

end.
